#   error.py  Flytec and Brauniger errors
#   Copyright (C) 2011  Tom Payne <twpayne@gmail.com>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.


class Error(RuntimeError):

    def __init__(self, msg=None):
        RuntimeError.__init__(self)
        self.msg = msg


class FlashError(Error):
    pass


class NotAvailableError(Error):
    pass


class TimeoutError(Error):
    pass


class ReadError(Error):
    pass


class WriteError(Error):
    pass


class ProtocolError(Error):
    pass
